use std::{
    env::{self, current_dir},
    ffi::OsStr,
    path::{Path, PathBuf},
    sync::{Arc, Mutex},
};

use git2::{
    build::{CheckoutBuilder, RepoBuilder},
    FetchOptions, RemoteCallbacks,
};

use crate::{
    config::{CuddleConfig, CuddleFetchPolicy},
    model::{CuddleBase, CuddlePlan},
};

#[derive(Debug, Clone, PartialEq)]
pub enum CuddleTreeType {
    Root,
    Leaf,
}

#[derive(Debug)]
pub struct CuddleContext {
    pub plan: CuddlePlan,
    pub path: PathBuf,
    pub node_type: CuddleTreeType,
}

pub fn extract_cuddle(
    config: CuddleConfig,
) -> anyhow::Result<Option<Arc<Mutex<Vec<CuddleContext>>>>> {
    let mut curr_dir = current_dir()?;
    curr_dir.push(".cuddle/");
    let fetch_policy = config.get_fetch_policy()?;
    if let CuddleFetchPolicy::Always = fetch_policy {
        if curr_dir.exists() {
            if let Err(res) = std::fs::remove_dir_all(curr_dir) {
                panic!("{}", res)
            }
        }
    }

    // Load main cuddle file.
    let cuddle_yaml = find_root_cuddle()?;
    if cuddle_yaml.is_none() {
        return Ok(None);
    }
    let cuddle_yaml = cuddle_yaml.unwrap();
    log::trace!(cuddle_yaml=log::as_debug!(cuddle_yaml); "Find root cuddle");

    let cuddle_plan = serde_yaml::from_str::<CuddlePlan>(cuddle_yaml.as_str())?;
    log::debug!(cuddle_plan=log::as_debug!(cuddle_yaml); "parse cuddle plan");

    let context: Arc<Mutex<Vec<CuddleContext>>> = Arc::new(Mutex::new(Vec::new()));
    context.lock().unwrap().push(CuddleContext {
        plan: cuddle_plan.clone(),
        path: current_dir()?,
        node_type: CuddleTreeType::Root,
    });

    // pull parent plan and execute recursive descent
    match cuddle_plan.base {
        CuddleBase::Bool(true) => {
            return Err(anyhow::anyhow!(
                "plan cannot be enabled without specifying a plan"
            ))
        }
        CuddleBase::Bool(false) => {
            log::debug!("plan is root: skipping");
        }
        CuddleBase::String(parent_plan) => {
            let destination_path = create_cuddle_local()?;
            let mut cuddle_dest = destination_path.clone();
            cuddle_dest.push("base");

            if !cuddle_dest.exists() {
                pull_parent_cuddle_into_local(parent_plan, cuddle_dest.clone())?;
            }

            recurse_parent(cuddle_dest, context.clone())?;
        }
    }

    Ok(Some(context))
}

fn create_cuddle_local() -> anyhow::Result<PathBuf> {
    let mut curr_dir = current_dir()?.clone();
    curr_dir.push(".cuddle/");

    if curr_dir.exists() {
        log::debug!(".cuddle/ already exists: skipping");
        return Ok(curr_dir);
    }

    std::fs::create_dir(curr_dir.clone())?;

    Ok(curr_dir)
}

fn create_cuddle(path: PathBuf) -> anyhow::Result<PathBuf> {
    let mut curr_dir = path.clone();
    curr_dir.push(".cuddle/");

    if curr_dir.exists() {
        log::debug!(".cuddle/ already exists: skipping");
        return Ok(curr_dir);
    }

    std::fs::create_dir(curr_dir.clone())?;

    Ok(curr_dir)
}

fn pull_parent_cuddle_into_local(
    parent_cuddle: String,
    destination: PathBuf,
) -> anyhow::Result<()> {
    let mut rc = RemoteCallbacks::new();
    rc.credentials(|_url, username_from_url, _allowed_types| {
        if *"true" == std::env::var("CUDDLE_SSH_AGENT").ok().unwrap_or("".into()) {
            git2::Cred::ssh_key_from_agent(username_from_url.unwrap())
        } else {
            git2::Cred::ssh_key(
                username_from_url.unwrap(),
                None,
                Path::new(&format!("{}/.ssh/id_ed25519", env::var("HOME").unwrap())),
                None,
            )
        }
    });

    rc.certificate_check(|_cert, _something| Ok(git2::CertificateCheckStatus::CertificateOk));

    let mut fo = FetchOptions::new();
    fo.remote_callbacks(rc);

    let co = CheckoutBuilder::new();
    RepoBuilder::new()
        .fetch_options(fo)
        .with_checkout(co)
        .clone(&parent_cuddle, &destination)?;

    log::debug!(parent_cuddle=log::as_display!(parent_cuddle); "pulled repository");

    Ok(())
}

fn recurse_parent(
    path: PathBuf,
    context: Arc<Mutex<Vec<CuddleContext>>>,
) -> anyhow::Result<Option<()>> {
    let cuddle_contents = find_cuddle(path.clone())?;
    if cuddle_contents.is_none() {
        return Ok(None);
    }
    let cuddle_plan = serde_yaml::from_str::<CuddlePlan>(&cuddle_contents.unwrap())?;

    let ctx = context.clone();
    if let Ok(mut ctxs) = ctx.lock() {
        ctxs.push(CuddleContext {
            plan: cuddle_plan.clone(),
            path: path.clone(),
            node_type: CuddleTreeType::Leaf,
        });
    } else {
        return Err(anyhow::anyhow!("Could not acquire lock, aborting"));
    }

    match cuddle_plan.base {
        CuddleBase::Bool(true) => Err(anyhow::anyhow!(
            "plan cannot be enabled without specifying a plan"
        )),
        CuddleBase::Bool(false) => {
            log::debug!("plan is root: finishing up");
            Ok(Some(()))
        }
        CuddleBase::String(parent_plan) => {
            let destination_path = create_cuddle(path.clone())?;
            let mut cuddle_dest = destination_path.clone();
            cuddle_dest.push("base");

            if !cuddle_dest.exists() {
                pull_parent_cuddle_into_local(parent_plan, cuddle_dest.clone())?;
            }
            recurse_parent(cuddle_dest, context.clone())
        }
    }
}

fn find_root_cuddle() -> anyhow::Result<Option<String>> {
    // TODO: Make recursive towards root
    let current_dir = env::current_dir()?;
    find_cuddle(current_dir)
}

fn find_cuddle(path: PathBuf) -> anyhow::Result<Option<String>> {
    for entry in std::fs::read_dir(path)? {
        let entry = entry?;
        let path = entry.path();

        let metadata = std::fs::metadata(&path)?;
        if metadata.is_file() && path.file_name().unwrap() == OsStr::new("cuddle.yaml") {
            return Ok(Some(std::fs::read_to_string(path)?));
        }
    }

    Ok(None)
}
